#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include "autoant.h"

uint16_t get_milliseconds(void) {

  struct timeval tv;
  gettimeofday(&tv,NULL);

  uint64_t t=tv.tv_sec*1000;
  t+=tv.tv_usec/1000;

  return t & 0xffff;
}


FILE *fd;
int ant_fd;

int16_t read_byte(void) { return (int16_t)fgetc(fd); }
void myseek(int offset) { fseek(fd, offset, SEEK_SET); }

void send_ant(uint8_t *c) {

    int i;
    int l=autoant_antlen(c);
    printf(" Sending Ant ");
    for (i=0; i<l; i++) {
        printf("%02x ",c[i]);
    }
    printf("\n");

    write(ant_fd, c, l);
    write(ant_fd, "\0\0", 3);
    tcdrain(ant_fd);
}

#define ANT_MAX_LENGTH (30)

#define ANT_BUF_LEN 32
uint8_t ant_rx_buf[ANT_BUF_LEN];

int ant_get_char(int64_t milliseconds_timeout) {
  
  char c;

  fd_set rfds;
  
  FD_ZERO(&rfds);
  FD_SET(ant_fd, &rfds);

  struct timeval tv = { milliseconds_timeout / 1000, (milliseconds_timeout * 1000) % 1000000 };

  if (1==select(1+ant_fd, &rfds, NULL, NULL, &tv)) {
    read(ant_fd, &c, 1);
    return c;
  } else { // timeout or select() error
    return -1;
  }
}

#define ANT_SYNC_BYTE (0xa4)

uint8_t *receive_ant(int timeout) {
  enum States {ST_WAIT_FOR_SYNC, ST_GET_LENGTH, ST_GET_MESSAGE_ID, ST_GET_DATA, ST_VALIDATE_PACKET};
  static enum States state = ST_WAIT_FOR_SYNC;

  //  timeout += 200;

  //printf("Waiting for message %.3f sec\n",(float)timeout/1000.0);

  uint16_t t0 = get_milliseconds();
  static int length;
  static uint8_t checksum;
  static int chars_in_buf=0;
  
  while (1) {
    int16_t time_remaining = t0 + timeout - get_milliseconds(); 
    if (time_remaining < 0) return NULL;

    int c=ant_get_char(time_remaining);
    if (c==-1)
      return NULL;
    
    c &= 0xff;

    switch (state) {
    case ST_WAIT_FOR_SYNC:
      if (c == ANT_SYNC_BYTE) {
	state = ST_GET_LENGTH;
	ant_rx_buf[chars_in_buf++] = c;
	checksum = ANT_SYNC_BYTE;

	//printf("SYNC\n");
      }
      break;
      
    case ST_GET_LENGTH:
      if ((c == 0) || (c > ANT_MAX_LENGTH)) {
	chars_in_buf=0;
	state = ST_WAIT_FOR_SYNC;
      }
      else {
	ant_rx_buf[chars_in_buf++] = c;
	checksum ^= c;
	length = chars_in_buf+c+1;
	state = ST_GET_MESSAGE_ID;
	//printf("LEN %d\n",c);
      }
      break;
      
    case ST_GET_MESSAGE_ID:
      ant_rx_buf[chars_in_buf++] = c;
      checksum ^= c;
      state = ST_GET_DATA;
      //printf("ID %x\n",c);
      break;
	
    case ST_GET_DATA:
      ant_rx_buf[chars_in_buf++] = c;
      checksum ^= c;
      if (chars_in_buf >= length){
	state = ST_VALIDATE_PACKET;
	//printf("DATADONE\n");
      }
      break;
      
    case ST_VALIDATE_PACKET:
      ant_rx_buf[chars_in_buf++] = c;
      //printf ("CHK %x %x\n",c,checksum);
      if (checksum == c){	

        int i;
        printf("Received Ant ");
        for (i=0; i<chars_in_buf; i++) {
            printf("%02x ",ant_rx_buf[i]);
        }
        printf("\n");
       
	chars_in_buf=0; // ready for next time
	state = ST_WAIT_FOR_SYNC;
	return ant_rx_buf;	
      }
      // chksum fail. start looking for message again
      chars_in_buf=0;
      state = ST_WAIT_FOR_SYNC;
      break;
    }  
  }
}


void write_message(char *message) {
    fprintf(stdout,"%s",message);
}

void warn_message(int warnno, int linenumber, int pos) {
    printf("At line %d byte %d : #%d (%s)\n",
           linenumber, pos, warnno,
           autoant_interpret_warning(warnno));
}

int setdtr (int on)
{
  int controlbits = TIOCM_DTR;
  if(on)
    return(ioctl(ant_fd, TIOCMBIC, &controlbits));
  else
    return(ioctl(ant_fd, TIOCMBIS, &controlbits));
} 

const int ant_baudrate=115200;
const char ant_filename[]="/dev/ttyANT2";

void init_tty(const char *dev) {
	struct termios tty_conf;
	
	ant_fd = open(dev, O_RDWR | O_NOCTTY | O_NONBLOCK);

	if(ant_fd < 0) {
	  fprintf(stderr,"Error opening %s", dev);
	  perror(dev);
	  exit(ant_fd);
	}

	tcgetattr(ant_fd, &tty_conf);

	tty_conf.c_cflag |= CLOCAL | CREAD;
	tty_conf.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL | ISIG | IEXTEN);
	tty_conf.c_lflag &= ~(OPOST);
	tty_conf.c_iflag &= ~(INLCR|IGNCR|ICRNL|IGNBRK);
	
	// stopbits
	tty_conf.c_cflag &= ~CSTOPB;

#define TRY(sp)						\
	if (ant_baudrate==sp) {		\
	  cfsetispeed(&tty_conf, B ## sp);		\
	  cfsetospeed(&tty_conf, B ## sp);		\
	}
	TRY(4800)
	else TRY(9600)
	else TRY(19200)
	else TRY(38400)
	else TRY(57600)
	else TRY(115200)
	else {
	  fprintf(stderr, "Invalid baudrate: %d\n",ant_baudrate);
	  exit(-1);
	}
#undef TRY
	
	//cfmakeraw(&tty_conf);
	
	tty_conf.c_cc[VMIN]=0;
   	tty_conf.c_cc[VTIME]=0;
	
	// Set the new options for the port...
	//tcflush(ant_fd, TCIFLUSH);
	tcsetattr(ant_fd, TCSANOW, &tty_conf);

	setdtr(0);
	setdtr(1);

}


autoant_ops_t ops={
  .receive_ant_message=receive_ant,
  .get_char=read_byte,
  .seek=myseek,
  .send_ant_message=send_ant,
  .write=write_message,
  .warn=warn_message,
  .get_milliseconds=get_milliseconds
};

int main(int argc, char *argv[]) {

  switch (argc) {
  case 1:
  case 2:
    printf("%s -- play autoant scripts through POSIX serial\n",argv[0]);
    printf("\nUse like this:\n");
    printf("   %s /dev/ttyUSB0 scriptname.ants\n",argv[0]);
    printf("\nSend bugs to mark@quarq.us\n\n");
    break;
  case 3:
    init_tty(argv[1]);
    fd=fopen(argv[2],"r");
    if (!fd) { perror("Opening file"); return errno; }
    return autoant_exec(&ops);
    break;
  default:
    printf("too many args\n");
  }
  return -1;
}
